/*
** Module   : ChuiWidgetFactory.java
** Abstract :
**
** Copyright (c) 2011-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111221 Created initial version
** 002 CA  20130129 Import change.
** 003 MAG 20140410 Add RECTANGLE widget creation.
** 004 EVL 20140610 Added TOGGLE-BOX widget support.
** 005 MAG 20140521 Refactored to implement support for GUI.
** 006 CA  20140711 Added MESSAGE-AREA widget support.
** 007 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 008 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 009 MAG 20141009 Deleted createMessageLine.
** 010 MAG 20141021 Renamed FrameImpl to FrameChuiImpl.
** 011 EVL 20141118 Adding scroll container creation method.
** 012 VIG 20150116 Added Menu classes mapping.
** 013 CA  20150323 Labels are always created using specific APIs; these APIs require the frame to
**                  which the label will belong, initially.
** 014 EVL 20150428 Adding drop-down container creation.
** 015 CA  20150509 Change to allow any widget to be contained by a scroll container.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Text;
import com.goldencode.p2j.ui.client.chui.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * CHUI-specific widget factory implementation.
 * <p>
 * The return type for the {@code create} APIs is different from the {@link WidgetFactory} 
 * versions because:
 * <ul>
 *    <li>it avoids adding at the signature the {@link ChuiOutputManager} type, i.e. instead
 *        of using {@code Label&lt;ChuiOutputManager&gt;} as return type, {@link LabelImpl}
 *        can be used directly, with the same effect. </li>
 *    <li> adds an additional protection against incorrect usage of the {@link WidgetFactory} 
 *    implementation classes: if the {@link WidgetFactory} implementation class is used 
 *    explicitly, it prevents accidental mixing of ChUI and GUI widgets (i.e. if used from 
 *    incorrect location). </li>
 * </ul> 
 */
public class ChuiWidgetFactory
extends WidgetFactoryAdapter<ChuiOutputManager>
{
   static 
   {
      classMap.put(AlertBox.class, AlertBoxImpl.class);
      classMap.put(Browse.class, BrowseImpl.class);
      classMap.put(BrowseColumn.class, BrowseColumn.class);
      classMap.put(Button.class, ButtonImpl.class);
      classMap.put(ComboBox.class, ComboBoxImpl.class);
      classMap.put(Editor.class, EditorImpl.class);
      classMap.put(FillIn.class, FillInImpl.class);
      classMap.put(Frame.class, FrameChuiImpl.class);
      classMap.put(Menu.class, MenuChuiImpl.class);
      classMap.put(MenuItem.class, MenuItemChuiImpl.class);
      classMap.put(SubMenu.class, SubMenuChuiImpl.class);
      classMap.put(RadioButton.class, RadioButtonImpl.class);
      classMap.put(RadioSet.class, RadioSet.class);
      classMap.put(Rect.class, RectangleImpl.class);
      classMap.put(ScrollableList.class, ScrollableListImpl.class);
      classMap.put(ScrollPane.class, ScrollPaneImpl.class);
      classMap.put(SelectionListBody.class, SelectionListBodyImpl.class);
      classMap.put(SelectionList.class, SelectionListImpl.class);
      classMap.put(Separator.class, SeparatorImpl.class);
      classMap.put(Skip.class, SkipImpl.class);
      classMap.put(Space.class, Space.class);
      classMap.put(StatusLine.class, StatusLineImpl.class);
      classMap.put(Text.class, TextImpl.class);
      classMap.put(ToggleBox.class, ToggleBoxImpl.class);
      classMap.put(Window.class, WindowChuiImpl.class);      
   }
   
   /**
    * Create new {@link Label} instance for given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Label text. 
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    *
    * @return  new {@link Label} instance.
    */
   @Override
   public LabelImpl createLabel(WidgetId id, String text, int frameId)
   {
      return new LabelImpl(id, text, frameId);
   }

   /**
    * Create new button for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          Button text.
    *          
    * @return  new {@link Button} instance.
    */
   @Override
   public ButtonImpl createButton(WidgetId id, Container<ChuiOutputManager> container, String text)
   {
      return new ButtonImpl(id, container, text);
   }

   /**
    * Create new toggle-box for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          Toggle Box text.
    *          
    * @return  new {@link ToggleBox} instance.
    */
   @Override
   public ToggleBoxImpl createToggleBox(WidgetId id, Container<ChuiOutputManager> container, String text)
   {
      return new ToggleBoxImpl(id, container, text);      
   }

   /**
    * Create new rectangle for given container.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    *          
    * @return  new {@link Rect} instance.
    */
   @Override
   public RectangleImpl createRectangle(WidgetId id, Container<ChuiOutputManager> container)
   {
      return new RectangleImpl(id, container);
   }

   /**
    * Create UI-specific {@link DropDown} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    *          
    * @return  instance of UI-specific implementation of {@link DropDown}.
    */
   @Override
   public DropDown<ChuiOutputManager> createDropDown(ComboBox<ChuiOutputManager> parent)
   {
      return new DropDown<ChuiOutputManager>(parent);
   }

   /**
    * Create UI-specific {@link ScrollableList} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    * @param   model
    *          Scrollable list model.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollableList}.
    */
   @Override
   public ScrollableListImpl createScrollableList(ComboBox<ChuiOutputManager> parent, 
                                                  ListModel<String>           model)
   {
      return new ScrollableListImpl(parent, parent.model());
   }

   /**
    * Create UI-specific {@link ScrollContainer} implementation.
    * 
    * @param   isScrollable
    *          The boolean flag that indicates whether the ScrollContainer has
    *          the scrolling capability.
    * @param   widget
    *          The widget being scrolled.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollPane}.
    */
   @Override
   public ScrollContainer createScrollContainer(boolean isScrollable, ScrollableWidget widget)
   {
      return new ScrollContainer(isScrollable, widget);
   }

   /**
    * Create UI-specific {@link ScrollPane} implementation.
    * 
    * @param   id
    *          The widget's ID.
    * @param   widget 
    *          Widget which will be scrolled.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollPane}.
    */
   @Override
   public ScrollPaneImpl createScrollPane(WidgetId id, Widget<ChuiOutputManager> widget)
   {
      return new ScrollPaneImpl(id, widget);
   }

   /**
    * Create UI-specific {@link Label} instance as a copy of given {@link Label}.
    * 
    * @param   id
    *          The widget's ID.
    * @param   labelInstance
    *          Source {@link Label} instance.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    *          
    * @return  instance of UI-specific implementation of {@link Label}.
    */
   @Override
   public LabelImpl createLabel(WidgetId id, Label<ChuiOutputManager> labelInstance, int frameId)
   {
      return new LabelImpl(id, labelInstance, frameId);
   }

   /**
    * Create UI-specific {@link RadioButton} instance.
    * 
    * @param   config
    *          Parent config. 
    * @param   label
    *          Radio button label. 
    * @param   value
    *          Radio button data type. 
    * 
    * @return  instance of UI-specific implementation of {@link RadioButton}.
    */
   @Override
   public RadioButtonImpl createRadioButton(RadioSetConfig config, 
                                            String         label, 
                                            BaseDataType   value)
   {
      return new RadioButtonImpl(config, label, value);
   }

   /**
    * Create UI-specific {@link Window} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link Window}.
    */
   @Override
   public WindowChuiImpl createWindow(WidgetId id)
   {
      return new WindowChuiImpl(id);
   }

   /**
    * Create UI-specific {@link StatusLine} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link StatusLine}.
    */
   @Override
   public StatusLineImpl createStatusLine(WidgetId id)
   {
      return new StatusLineImpl(id);
   }
   
   /**
    * Create UI-specific {@link MessageArea} instance.
    * 
    * @param   id
    *          The widget's ID.
    * @param   lines
    *          Number of visible lines in message area.         
    * 
    * @return  instance of UI-specific implementation of {@link MessageArea}.
    */
   @Override
   public MessageAreaImpl createMessageArea(WidgetId id, int lines)
   {
      return new MessageAreaImpl(id, lines );
   }
}
