/* Module   : ColumnLayout.java
** Abstract : Column LayoutManager implementation for CHUI
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050623   @21550 Creation initial version.
** 002 EVL 20050627   @21557 Adding handling of Space - one more artificial
**                           widget.
** 003 EVL 20050729   @21915 Removing the ugly tabulation from source.
** 004 EVL 20050731   @21925 Fixing the Null pointer exception in the case
**                           when the label is null.
** 005 SIY 20050831   @22528 Formatting and spelling fixes. 
** 006 SIY 20060207   @24483 Updated to support other widget types.
** 007 SIY 20060408   @25433 Rewritten to match widget placement algorithm
**                           used by Progress.
** 008 SIY 20060412   @25600 Removed unused code/variables/constants.
** 009 SIY 20080702   @39035 Fixed handling of COLON and fixed handling of 
**                           layouts with column count greater than 4.
** 010 SIY 20100619          Switching to own implementation of character UI.
** 011 SIY 20100804          Changes in imports, some refactorings.
** 012 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 013 SIY 20111122          Intermediate version before introduction of factory.
** 014 SIY 20111221          Switched to widget factory.
** 015 CA  20140805          Added generics to force the abstract widget classes to not use 
**                           implementation-specific code: the abstract widgets will have access 
**                           only to the base APIs provided by the OutputManager.
** 016 HC  20140911          GUI coordinate support - data types refactoring.
** 017 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 018 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 019 CA  20150114          Side-labels must be linked with their corresponding widget's location
**                           in the down's frame body.
** 020 CA  20150323          Pass the frame ID when creating a label (to properly resolve the 
**                           font in GUI).
** 021 HC  20150323          Removed business logic from config classes.
** 022 CA  20150406          A minor change in computing the label's width.  This is still 
**                           incomplete for GUI support.
** 023 CA  20150509          Change to allow any widget to be contained by a scroll container.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This is the LayoutManager which places the widgets into specified number
 * of the columns and put the container into SIDE-LABELS mode.
 */
public class ColumnLayout<O extends OutputManager<?>>
implements LayoutManager<O>
{
   /** The number of columns requested to widget representation. */
   private int columns = 1;

   /** This value is dependent from the alignment type. */
   private int alignParam = -1;

   /** This field tracks container's NO-LABELS attribute. */
   private boolean labels = false;
   
   /** This field tracks container's ATTR-SPACE attribute. */
   private boolean attr_space = false;
   
   /**
    * Creates the new instance of the LayoutManager with the given number of
    * the columns. It is also possible to declare the alignment, type and the
    * parameter. The following alignments are implemented:
    * <p>
    * <ul>
    * <li>ALIGN_DEFAULT means the widgets will be aligned on the leftmost
    * column position. The value of the align parameter is meaningless
    * <li>ALIGN_COLON means usage of the colon char positioning. The align
    * parameter is the column offset of the colon char for the widget within
    * the current column of the widgets
    * <li>ALIGN_TO means the widgets will be aligned on the leftmost column
    * position. The align parameter is the column offset of the rightmost
    * widget position within the current column of the widgets
    * </ul>
    * <p>
    * 
    * @param   columns
    *          The number of the columns to be presented inside the container.
    */
   public ColumnLayout(int columns)
   {
      if (columns > 1)
         this.columns = columns;

      this.alignParam = 14;
   }
   
   /**
    * Makes the widgets coordinates recalculation based on the current columns
    * requirement.
    * 
    * @param   container
    *          The container to which this LayoutManager should be applied.
    */
   @Override
   public void doLayout(Container<O> container)
   {
      calcLayout(container, true);
   }

   /**
    * Calculates the minimum size of the container which will be big enough to
    * hold all widgets.
    * 
    * @param   container
    *          The container which size should be calculated.
    * 
    * @return  The size of the container which is big enough to hold all
    *          widgets.
    */
   @Override
   public Dimension minimumSize(Container<O> container)
   {
      return calcLayout(container, false);
   }

   /**
    * Perform actual layout calculations and placement.
    * 
    * @param   container
    *          Container to process.
    * @param   doPlacement
    *          <code>true</code> place widgets, <code>false</code>
    *          simulate placement.
    * 
    * @return  Containing rectangle size.
    */
   @SuppressWarnings("unchecked")
   private Dimension calcLayout(Container<O> container, boolean doPlacement)
   {
      if (container instanceof ScrollContainer)
      {
         ScrollContainer<?> sc = (ScrollContainer<?>) container;
         Frame<O> frame = (Frame<O>) sc.getWidget();
         labels = frame.isLabels();
         attr_space = frame.isAttrSpace();
      }
      
      Dimension res = new Dimension(0, 0);
      Widget<O>[] c = container.widgets();
      
      int curRow = 0;
      int curCol = 0;
      int maxHeight = 0;
      int colCount = 0;
      
      // in chui char coordinates are always whole numbers
      int screenWidth = (int) container.window().screenWidth();
      int currColumns = columns;

      if ((screenWidth - 2)/currColumns < (alignParam + 2) && currColumns > 1)
         currColumns = (screenWidth - 2)/(alignParam + 2);
      
      int step = (screenWidth - 2)/currColumns;
      double dstep = (screenWidth - 2.0)/currColumns - 0.5;
      
      for (int i = 0; i < c.length; i++)
      {
         if (c[i] instanceof Skip)
         {
            curRow   += maxHeight + c[i].height();
            curCol    = 0;
            colCount  = 0;
            maxHeight = 0;
            continue;
         }
         
         if (c[i] instanceof Space)
         {
            curCol += c[i].width();
            continue;
         }
         
         Dimension dim = c[i].dimension();
         
         boolean hasLabel = false;
         LabeledWidget<O> lw = null;
         
         int colon = BaseConfig.INV_COORD;
         int alignLen = alignParam + (2/currColumns) * 2;
         
         if (labels && c[i] instanceof LabeledWidget)
         {
            lw = (LabeledWidget<O>) c[i];
            colon = (int) lw.getColon();
            
            if(lw.getLabelText() != null       && 
               lw.getLabelText().length() > 0  &&
               (labels || lw.isForceLabel())   && 
               !lw.isNoLabels()) 
            {
               int len = (int) ZeroColumnLayout.labelWidth(lw, lw.getLabelText() + ": ");
               
               if (colon != BaseConfig.INV_COORD)
                  alignLen = len;
                  
               dim.width += alignLen;

               hasLabel = true;
            }
         }
         
         int pos = curCol/step + (((curCol % step) > 0) ? 1:0);
         
         if (colon != BaseConfig.INV_COORD)
         {
            if (curCol > colon)
            {
               curRow += maxHeight;
               maxHeight = 0;
            }
            
            curCol = colon - alignLen + 2;
         }
         else
            curCol = (int)Math.round(pos*dstep);
         
         // in chui char coordinates are always whole numbers
         if ((curCol + (int) dim.width + 2) >= screenWidth)
         {
            curRow += maxHeight;
            curCol = 0;
            maxHeight = 0;
         }
         
         if (doPlacement)
         {
            // perform widget placement
            if (hasLabel && lw != null)
            {
               String text = lw.getLabelText();
               
               if (text.length() > (alignLen - 2))
                  text = text.substring(0, alignLen - 2);
               else
                  text = "                ".substring(0, 
                                                      alignLen - 
                                                      text.length() - 2) +
                                                      text;
               WidgetId nextId = c[i].getId().createId(WidgetId.nextID());
               Label<O> label = (Label<O>) container.screen().getFactory()
                                                    .createLabel(nextId,
                                                                 text + ": ", 
                                                                 c[i].config().frameId);
               label.setLocation(curCol, curRow);
               
               if (lw.getLabelInstance() != null)
               {
                  container.remove(lw.getLabelInstance());
               }
               
               lw.setLabelInstance(label);
               container.add(label);
               c[i].setLocation(curCol + alignLen, curRow);
            }
            else
            {
               c[i].setLocation(curCol, curRow);
            }
         }

         if (maxHeight < c[i].height())
            maxHeight = (int) c[i].height();
         
         curCol += (int) dim.width;
         colCount = curCol/step + (((curCol % step) > 0) ? 1:0);
         
         // in chui char coordinates are always whole numbers
         if ((curRow + maxHeight) > (int) res.height)
            res.height = curRow + maxHeight;
         
         if (curCol > (int) res.width)
            res.width = curCol;
         
         if (colCount >= currColumns)
         {
            curRow += maxHeight;
            curCol = 0;
            maxHeight = 0;
         }
      }
      
      if (attr_space)
         res.width += 1;
      
      return res;
   }
}
