/*
** Module   : TempContainer.java
** Abstract : Temporary container used for layout processing.
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100915 Created initial version
** 002 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 003 SIY 20111202 Updated imports.
** 004 SIY 20111221 Implemented screen() method.
** 005 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 HC  20150323 Removed business logic from config classes.
** 008 CA  20150429 Added NPE protection for screen().
** 009 CA  20150509 Allow scroll container for any widgets, not just frames.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.client.widget.*;

/**
 * Temporary container used for second and subsequent layout processing.
 * <p>
 * ENABLE processing may change order of the widgets in the container. Since
 * order also affects layout processing, original widget order is saved and
 * then all layout processing is done using this special container as a 
 * temporary holder for frame widgets. 
 */
public class TempContainer<O extends OutputManager<?>>
extends ScrollContainer<O>
{
   /** Source scroll container used to retrieve some useful info. */
   private ScrollContainer<O> base = null;
   
   /**
    * Construct an instance of container using given list of widgets and
    * dimension.
    * 
    * @param    widgets
    *           Array of <code>Widget</code> instances.
    * @param    base
    *           Base container for which this one will work as substitute.
    * @param    frame
    *           instance of <code>FrameHelper</code> for this frame
    */
   public TempContainer(Widget<O>[] widgets, ScrollContainer<O> base, Frame<O> frame)
   {
      super(false, frame);
      this.base = base;
      this.setParent(base.parent());
      setSize(base.dimension());
      
      // note that direct call is used instead of Component.add
      // to prevent changing of the reference to parent
      for (int i = 0; i < widgets.length; i++)
      {
         addDirectly(widgets[i]);
      }
      
      Widget<O>[] c = base.widgets();
      
      for (int i = 0; i < c.length; i++)
      {
         if (c[i] instanceof Label && ((Label<O>) c[i]).isDelimiter())
         {
            addDirectly(c[i]);
         }
      }
   }

   /**
    * Forwarder for add(Component) method.
    * 
    * @param    component
    * 
    * @return   <code>this</code> for fluent syntax support.
    */
   @Override
   public Container<O> add(Widget<O> component)
   {
      return base.add(component);
   }
   
   /**
    * Forwarder for remove(Component) method.
    * 
    * @param    component
    */
   @Override
   public void remove(Widget<O> component) 
   {
      base.remove(component);
   }
   
   /**
    * Extended implementation of replace method.
    * 
    * @param    oldc
    *           Existing component.
    * @param    newc
    *           Component which will replace old component.
    */
   @Override
   public void replace(Widget<O> oldc, Widget<O> newc)
   {
      // regular action
      base.replace(oldc, newc);
      
      // extension - changing non-body list
      ((Frame<O>) getWidget()).fixNonBody(oldc, newc);
   }
   
   /**
    * Get access to {@link OutputManager} instance.
    * 
    * @return  reference to {@link OutputManager}.
    */
   @Override
   public O screen()
   {
      return base == null ? null : base.screen();
   }
}
