/*
** Module   : EditorGuiImpl.java
** Abstract : EDITOR widget implemention. GUI-specific version.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20150509 Created initial version.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.MouseEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.ScrollPaneGuiImpl.ShowBars;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of Progress EDITOR component for GUI.
 */
public class EditorGuiImpl
extends Editor<GuiOutputManager>
implements ScrollableWidget
{
   /** Screen driver */
   private GuiDriver gd;

   /** Output manager */
   private GuiOutputManager screen;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;

   /** Font resolver */
   private GuiFontResolver gf;

   /** The scrollable container to place inside the ScrollPane. */
   private ScrollableContainer contentPane = null;
   
   /** Scroll pane which is actual container. */
   private ScrollPaneGuiImpl editorScroll = null;

   /** The editor content, where the text is actually drawn. */
   private final EditorContent editor = new EditorContent();
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public EditorGuiImpl(WidgetId id, EditorConfig cfg)
   {
      super(id, cfg);

      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // set frame default colors
      gc = new GuiColorResolver(gd, cfg, SysColor.COLOR_WINDOW, SysColor.COLOR_WINDOWTEXT);
      // box and title colors
      gc.boxColor     = gc.sysFgColor;
      // output manager
      screen = screen();
      cc = screen.coordinates();
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, EditorConfig wcfg)
   {
      super.initialize(id, wcfg);
      gf = new GuiFontResolver((Window<?>) ancestor(), gd, config(), false);

      WidgetFactory<GuiOutputManager> factory = screen().getFactory();

      contentPane = new ScrollableContainer(WidgetId.nextID());

      editorScroll = (ScrollPaneGuiImpl) factory.createScrollPane(WidgetId.nextID(), contentPane);
      add(editorScroll);
      
      editorScroll.config().sysbgcolor = SysColor.COLOR_WINDOW;
      // register scroll pane and its widgets for mouse events
      WidgetRegistry<?> reg = OutputManager.instance().getRegistry();
      reg.addWidget((Widget) editorScroll);
      editorScroll.registerWidgets((WidgetRegistry<GuiOutputManager>) reg);
      
      contentPane.add(editor);
   }
   
   /**
    * Set focus to first focusable widget.  For editors, to not allow the scrollbar buttons to
    * receive focus, will always return this widget.
    */
   @Override
   public void firstFocus()
   {
      setFocusInt(this);
   }
   
   /**
    * Process KeyEvents for this widget.
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      int key = ke.actionCode();

      Set<Integer> unsupportedEvents = new HashSet<>();
      // unsupportedEvents.add(Keyboard.KA_CURSOR_DOWN);
      // unsupportedEvents.add(Keyboard.KA_CURSOR_UP);
      // unsupportedEvents.add(Keyboard.KA_CURSOR_LEFT);
      // unsupportedEvents.add(Keyboard.KA_CURSOR_RIGHT);
      // unsupportedEvents.add(Keyboard.KA_HOME);
      unsupportedEvents.add(Keyboard.KA_LEFT_END);
      unsupportedEvents.add(Keyboard.KA_RIGHT_END);
      // unsupportedEvents.add(Keyboard.KA_BACKSPACE);
      // unsupportedEvents.add(Keyboard.KA_DELETE_CHARACTER);
      unsupportedEvents.add(Keyboard.KA_CLEAR);
      unsupportedEvents.add(Keyboard.KA_DELETE_END_LINE);
      unsupportedEvents.add(Keyboard.KA_DELETE_LINE);
      // unsupportedEvents.add(Keyboard.KA_END);
      unsupportedEvents.add(Keyboard.KA_DELETE_WORD);
      unsupportedEvents.add(Keyboard.KA_PREV_WORD);
      unsupportedEvents.add(Keyboard.KA_NEXT_WORD);
      unsupportedEvents.add(Keyboard.KA_PAGE_DOWN);
      unsupportedEvents.add(Keyboard.KA_PAGE_UP);
      unsupportedEvents.add(Keyboard.KA_BELL);

      if (unsupportedEvents.contains(key))
      {
         return;
      }
      
      if (key == Keyboard.KA_HOME)
      {
         cursorBeginOfLine();
         refresh();
         return;
      }
      else if (key == Keyboard.KA_END)
      {
         cursorEndOfLine();
         refresh();
         return;
      }

      super.processKeyEvent(ke);
   }
   
   /**
    * Sets the new variable value in the editor.
    * 
    * @param   value
    *          New value for the screen buffer.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      if (editorScroll != null)
      {
         resizeScrollPane();
      }

      super.setValue(value);
   }
   
   /**
    * Sets position of terminal's cursor unconditionally (even if this editor is not in focus).
    */
   @Override
   public void drawCaretUnconditionally()
   {
      if (!isEnabled())
      {
         return;
      }

      gd.setGuiFont(gf.font().font);

      NativePoint origin = new NativePoint(editor.screenPhysicalLocation());
      
      int cursorX = 0;
      
      Line l = getLine();
      if (l != null)
      {
         int absX = absoluteX();

         String txt = l.getText();
         txt = absX > 0 ? txt.substring(config.scrollX, absX) : "";
         
         // System.err.println(txt);
         cursorX = gd.getTextWidth(txt, gf.font().font);
      }
      
      origin = origin.offset(cursorX, config.cursorY * gf.font().fontHeight);

      final NativeDimension d = editor.physicalDimension();
      final NativeRectangle clip = clipRectangle(origin, d);

      gd.draw(origin, clip, new Runnable()
      {
         @Override
         public void run()
         {
            gd.setColor(gc.fgColor);
            gd.drawLine(0, 0, 0, gf.font().fontHeight);
         }
      });
   }

   /**
    * Render this editor.
    */
   @Override
   public void draw(Color color)
   {
      if (!isVisible())
      {
         return;         
      }

      // determine frame colors
      gc.colors();

      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      final NativeRectangle clip = clipRectangle(p, d);
      
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            gd.setGuiFont(gf.font().font);

            // clear the area
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, d.width, d.height);
            
            // draw box if defined
            if (config().box)
            {
               gd.setColor(gc.boxColor);
               gd.drawRect(0, 0, d.width, d.height);
            }

            editorScroll.draw();
         }
      });
   }

   /**
    * Destroy this widget.
    */
   @Override
   public void destroy()
   {
      super.destroy();
      
      if (contentPane != null)
      {
         contentPane.destroy();
      }
      
      if (editorScroll != null)
      {
         editorScroll.destroy();
      }
   }
   
   /**
    * Enable/disable the widget for editing.  During this operation, any data values are saved off 
    * for subsequent restore (during editing) using the <code>RECALL</code> key.
    * <p>
    * In GUI, it registers this widget with the driver, to allow responsive actions when the
    * mouse is hovered over the widget.
    * 
    * @param    enable
    *           <code>true</code> to enable editing in this widget.
    */
   @Override
   public void setEnabled(boolean enable)
   {
      boolean isEnabled = isEnabled();
      
      super.setEnabled(enable);
      
      int id = this.config.id.asInt();

      if (enable)
      {
         if (!isEnabled)
         {
            gd.registerEditableWidget(id);
         }
      }
      else
      {
         if (isEnabled)
         {
            gd.deregisterWidget(id);
         }
      }
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      
      if (!hasFocus())
      {
         ThinClient.getInstance().independentEventDrawingBracket(this, new Runnable()
         {
            @Override
            public void run()
            {
               // process the focus here, so we can continue with the special processing bellow.
               requestFocus();
            }
         });
      }
      
      NativePoint editorLoc = editor.screenPhysicalLocation();
      int nrow = e.getY() - editorLoc.y;
      int ncol = e.getX() - editorLoc.x;
      
      Window<GuiOutputManager> window = window();
      ScreenDriver<?> driver = screen().getInstanceDriver();
      int fontNum = gf.resolveFontNum();
      int nfontHeight = FontManager.getFontHeight(window, fontNum, driver);
      int editRow = nrow / nfontHeight;
      
      config.cursorY = Math.min(editRow, parsedText.size() - 1);

      Line line = getLine();
      if (line != null)
      {
         String txt = "";
         config.cursorX = line.length();
         int idx = 0;
         while (txt.length() < line.length())
         {
            int ntxtWidth = gd.getTextWidth(txt, gf.font().font);
            if (ntxtWidth >= ncol)
            {
               config.cursorX = txt.length();
               break;
            }
            
            idx = idx + 1;
            txt = line.getText().substring(0, idx);
         }
      }

      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();

      repaint();
   }

   /**
    * Mark this widget as being scrollable.
    * 
    * @param    isScrollable
    *           <code>true</code> to set this widget as scrollable.
    */
   @Override
   public void setScrollable(boolean isScrollable)
   {
      // nothing to do
   }

   /**
    * Check if this widget is scrollable.
    * 
    * @return   Always <code>true</code>
    */
   @Override
   public boolean isScrollable()
   {
      return true;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    *           The previous configuration.
    */
   @Override
   public void afterConfigUpdate(EditorConfig beforeUpdate)
   {
      super.afterConfigUpdate(beforeUpdate);

      if (editorScroll == null)
      {
         return;
      }
      
      contentPane.config().bgcolor = config.bgcolor;
      contentPane.config().fgcolor = config.fgcolor;
      editorScroll.config().bgcolor = config.bgcolor;
      editorScroll.config().fgcolor = config.fgcolor;

      boolean vbar = config.scrollbarVertical;
      boolean hbar = config.scrollbarHorizontal;
      boolean both = (vbar && hbar);
      editorScroll.setShowBars(hbar || vbar ? ShowBars.ALWAYS : ShowBars.NEVER);
      editorScroll.setScrollBars(both ? ScrollBar.Bars.BOTH 
                                      : hbar ? ScrollBar.Bars.HORIZONTAL 
                                             : vbar ? ScrollBar.Bars.VERTICAL 
                                                    : ScrollBar.Bars.NONE);
   
      WidgetRegistry<?> reg = OutputManager.instance().getRegistry();
      editorScroll.registerWidgets((WidgetRegistry<GuiOutputManager>) reg);
      
      if (gf == null)
      {
         // next, we initialize values after the widget has been realized
         return;
      }
      
      ScreenDriver<?> driver = screen().getInstanceDriver();
      
      int fontNum = gf.resolveFontNum();
      int nfontHeight = FontManager.getFontHeight(window(), fontNum, driver);
      int nfontWidth = FontManager.getFontWidth(window(), fontNum, driver);
      double fontWidth = screen().coordinates().widthFromNative(nfontWidth);
      double fontHeight = screen().coordinates().widthFromNative(nfontHeight);
            
      if (hbar)
      {
         editorScroll.hbar().setStep(fontWidth);
         editorScroll.hbar().setEnabled(false);
      }
      
      if (vbar)
      {
         editorScroll.vbar().setStep(fontHeight);
         editorScroll.vbar().setEnabled(false);
      }
      
      contentPane.addScrollListener(editor);
      
      resizeScrollPane();
   }

   /**
    * Refresh the editor.
   */
   @Override
   protected void refresh()
   {
      final NativePoint origin = screenPhysicalLocation();
      final NativeDimension d = physicalDimension();

      final NativeRectangle clip = clipRectangle(origin, d);

      gd.drawFromOrigin(origin, clip, new Runnable()
      {
         @Override
         public void run()
         {
            draw();
            requestSync();
         }
      });
   }
   
   /**
    * Check if the given text needs to output on a new line.
    * 
    * @param    subtext
    *           The text.
    *           
    * @return   <code>true</code> if the text will output on a new line.
    */
   @Override
   protected boolean goToNewLine(String subtext)
   {
      if (config.scrollbarHorizontal)
      {
         return false;
      }
      
      // get text width using the current font
      int txtWidth = gd.getTextWidth(subtext, gf.font().font);
      int editWidth = editor.physicalDimension().width;
      
      return txtWidth > editWidth;
   }
   
   /**
    * Split absolute horizontal cursor position into two components - position
    * of the first visible character and relative cursor position in the
    * visible region.
    * 
    * @param   pos
    *          The new absolute cursor position.
    */
   @Override
   protected void cursorPos(int pos)
   {
      Line line = getLine();
      config.cursorX = (line == null || line.length() == 0 ? 0 : pos % line.length());

      super.cursorPos(pos);
   }

   /**
    * Merge line with subsequent one(s) if possible
    * 
    * @param   row
    *          line's row
    */
   @Override
   protected void mergeLines(int row)
   {
      StringBuffer buff = new StringBuffer();
      for (Line line : parsedText)
      {
         buff.append(line.getText());
         if (line.isNewLine())
         {
            buff.append(LF);
         }
      }
      // System.out.println(buff.toString());
      // parsing the content forces the editor to show what it really holds - so any errors from
      // processing a key are immediately viewed
      // TODO: improve this
      parseContent(buff.toString(), -1, false);
      
      
      /*
      int edWidth = editor.physicalDimension().width;
      java.awt.Font font = gf.font().font;
      
      for(; row < parsedText.size() - 1; row++)
      {
         Line current = parsedText.get(row);
         if (current.isNewLine())
         {
            break;
         }
         
         int len = edWidth - gd.getTextWidth(current.getText(), font);
         if (!config.scrollbarHorizontal && len <= 0)
         {
            break;
         }
         
         Line next = parsedText.get(row + 1);
         String s = next.getText();
         int nextLen = gd.getTextWidth(s, font);
         
         if (config.scrollbarHorizontal || nextLen <= len)
         {
            parsedText.set(row, 
                           new Line(current.getText() + s, current.getOffset(), next.isNewLine()));
            parsedText.remove(row + 1);
            break;
         }
         
         int pos = s.lastIndexOf(' ', len);
         if (pos < 0)
         {
            break;
         }
         
         parsedText.set(row, new Line(current.getText() + s.substring(0, pos + 1), 
                                      current.getOffset(), 
                                      false));
         
         parsedText.set(row + 1, new Line(s.substring(pos + 1), 
                                          current.getOffset() + pos, 
                                          next.isNewLine()));
      }
      */
   }
   
   /**
    * Get the next split position of this line, so that it fits the editor.
    * 
    * @param    line
    *           The line to be split.
    * @param    splitAfter
    *            minimal split position.
    *           
    * @return   An index where the split ends.
    */
   @Override
   protected int getNextSplitPosition(String line, int splitAfter)
   {
      int pos = line.lastIndexOf(SPACE, line.length());
      if (pos < splitAfter)
      {
         pos = line.length() - 1;
      }
      
      return pos;
   }
   
   /**
    * Sets the cursor position in the current line
    *
    * @param    pos
    *           the new cursor position value\
    * @param    redraw
    *           request redraw/sync
    *           
    * @return   <code>true</code> if the cursor position is valid and set 
    */
   @Override
   public boolean setCursorChar(int pos, boolean redraw)
   {
      int lineLength = getLine().length();
      if (pos < 0 && pos > lineLength)
      {
         return false;
      }
      
      if (pos < lineLength - 1)
      {
         config.cursorX = Math.min(pos, lineLength);
         config.scrollX = 0;
      }
      else
      {
         config.cursorX = (lineLength == 0 ? 0 : pos % lineLength);
         config.scrollX = pos - config.cursorX;
      }

      return super.setCursorChar(pos, redraw);
   }
   
   /**
    * Move cursor one char right.
    */
   @Override
   protected void cursorRight()
   {
      Line line = getLine();
      int absX = absoluteX();
      String txt = (line == null ? "" : line.getText());
      String visTxt = txt.substring(config.scrollX, Math.min(absX + 1, txt.length()));
      int ntxtLength = txt.length() == 0 ? 0 : gd.getTextWidth(visTxt, gf.font().font);
      int edWidth = editor.physicalDimension().width;
      
      if (config.scrollX + config.cursorX + 1 > txt.length())
      {
         // just go to next row, nothing else to do
         if (absoluteY() < parsedText.size() - 1)
         {
            // go to next line
            cursorDown();
            cursorBeginOfLine();
         }
         
         return;
      }
         
      if (txt.length() > 0)
      {
         // if the current text is still visible, just advance the cursor
         if (ntxtLength < edWidth)
         {
            // just advance the cursor; scrollX remains where it is
            config.cursorX++;
         }
         else
         {
            // the width no longer matches. we need to right-anchor the text on the current 
            // absX + 1 and find a new text which fits the screen
            int len = absX;
            String newVisTxt = null;
            while (len >= 0)
            {
               newVisTxt = txt.substring(len, absX + 1);

               int visLength = newVisTxt.length() == 0 ? 0 
                                                       : gd.getTextWidth(newVisTxt, gf.font().font);
               if (visLength > edWidth)
               {
                  // found one
                  newVisTxt = newVisTxt.substring(1);
                  break;
               }
               
               len = len - 1;
            }
            
            config.scrollX = absX - newVisTxt.length() + 1;
            config.cursorX = newVisTxt.length();
         }
      }
      
      super.cursorRight();
   }
   
   /**
    * Move cursor one row down, if possible.
    */
   @Override
   protected void cursorDown()
   {
      super.cursorDown();
      
      int absoluteX = absoluteX();
      Line line = getLine();
      if (line == null || absoluteX >= line.length())
      {
         cursorEndOfLine();
      }
   }

   /**
    * Move cursor one row up, if possible.
    */
   @Override
   protected void cursorUp()
   {
      super.cursorUp();
      
      int absoluteX = absoluteX();
      Line line = getLine();
      if (line == null || absoluteX >= line.length())
      {
         cursorEndOfLine();
      }
   }
   
   /**
    * Move cursor one char left, if possible.
    */
   @Override
   protected void cursorLeft()
   {
      if (config.cursorX + config.scrollX == 0 && absoluteY() > 0)
      {
         cursorUp();
         cursorEndOfLine();
      }
      else if (config.cursorX > 0)
      {
         config.cursorX--;
      }
      else
      {
         if (config.scrollX > 0)
            config.scrollX--;
      }

      super.cursorLeft();
   }

   /**
    * Move cursor to the end of line.
    */
   @Override
   protected void cursorEndOfLine()
   {
      int len = getLine().length() - 1;
      String line = getLine().getText();
      
      String txt = "";
      while (len >= 0)
      {
         txt = line.substring(len);
         
         int nwidth = gd.getTextWidth(txt, gf.font().font);
         if (nwidth >= editor.physicalDimension().width)
         {
            txt = txt.substring(1);
            break;
         }
         
         len = len - 1;
      }

      if (txt.length() == 0)
      {
         config.scrollX = 0;
         config.cursorX = 0;
      }
      else if (txt.length() == line.length())
      {
         config.scrollX = 0;
         config.cursorX = txt.length();
      }
      else
      {
         config.scrollX = getLine().length() - txt.length();
         config.cursorX = txt.length();
      }
      
      super.cursorEndOfLine();
   }

   /**
    * Handle BACKSPACE key.
    */
   @Override
   protected void backspace()
   {
      Line line = getLine();
      boolean merge = (content.length() > 0 && line != null && config.cursorX == 0 && absoluteY() > 0);
      
      if (merge)
      {
         // simulate going up, so we can save the cursorX/scrollX
         cursorEndOfPrevLine();
         int cursorX = config.cursorX;
         int scrollX = config.scrollX;
         
         // go back to our line
         cursorDown();
         cursorBeginOfLine();
         
         // delete the NL for the previous line
         StringBuffer buff = new StringBuffer(content);
         buff.deleteCharAt(line.getOffset() - 1);
         parsedText.get(absoluteY() - 1).setNewLine(false);
         
         setContent(buff);
         
         mergeLines(absoluteY() - 1);
         
         // move the cursor
         if (config.cursorY == 0)
         {
            config.scrollY--;
         }
         else
         {
            config.cursorY--;
         }
         
         config.cursorX = cursorX;
         config.scrollX = scrollX;
      }
      else
      {
         super.backspace();
      }
      
      // System.out.println(content);
   }
   
   /**
    * Delete character at current cursor position - called by {@link #processKeyEvent}.
    *
    * @return   <code>true</code> if something was deleted.
    */
   @Override
   protected void deleteCharacterKey()
   {
      Line line = getLine();
      boolean merge = (content.length() > 0         && 
                       line != null                 && 
                       absoluteX() == line.length() && 
                       absoluteY() < parsedText.size());
      
      if (merge)
      {
         if (content.length() > line.getOffset() + line.length() && 
             (line.length() != 0 || absoluteY() != parsedText.size() - 1))
         {
            // delete the NL for the current line
            StringBuffer buff = new StringBuffer(content);
            buff.deleteCharAt(line.getOffset() + line.length());
            line.setNewLine(false);   
            setContent(buff);
   
            mergeLines(absoluteY());
         }
      }
      else
      {
         super.deleteCharacter();
      }
      
      // System.out.println(content);
   }

   /**
    * Remove trailing whitespaces. In GUI, this is a no-op.
    * 
    * @param   value
    *          the string to be truncated
    *    
    * @return  the truncated argument
    */
   protected String rtrim(String value)
   {
      return value;
   }

   /**
    * Get absolute horizontal position in text.
    * 
    * @return  The horizontal position in the text.
    */
   protected int absoluteX()
   {
      // must not exceed line length
      Line line = getLine();
      return Math.min(config.cursorX + config.scrollX, line == null ? 0 : line.length()); 
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * <p>
    * For editors, if the sources is either {@link #editor} or {@link #contentPane}, then return
    * this widget.
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      Widget<GuiOutputManager> src = super.findMouseSource(p);
      
      return (src == editor || src == contentPane) ? this : src;
   }

   /**
    * Resize editor scroll pane.
    */
   private void resizeScrollPane()
   {
      boolean box = config().box;
   
      // location
      NativePoint np = new NativePoint(0, 0);
      if (box)
      {
         np.translate(1, 1);
      }
   
      editorScroll.setPhysicalLocation(np.x, np.y);
      editorScroll.setLocation(cc.pointFromNative(np));

      editor.setPhysicalLocation(np.x + 2, np.y + 2);
      editor.setLocation(cc.pointFromNative(np));

      // size
      NativeDimension nd = physicalDimension();
      nd.width  = nd.width - (box ? 2 : 0);
      nd.height = nd.height - (box ? 2 : 0);
      editorScroll.setSize(cc.dimensionFromNative(nd));
      editorScroll.adjustViewportSize();

      ScreenDriver<?> driver = screen().getInstanceDriver();
      Window<GuiOutputManager> window = window();
      CoordinatesConversion cc = screen().coordinates();
      
      int fontNum = gf.resolveFontNum();
      int nfontHeight = FontManager.getFontHeight(window, fontNum, driver);
      int nfontWidth = FontManager.getFontWidth(window, fontNum, driver);
      double fontHeight = cc.heightFromNative(nfontHeight);
      double fontWidth = cc.widthFromNative(nfontWidth);

      double maxLineWidth = maxLineWidth();
      
      // compute the inner-lines/-chars
      NativeDimension dim = editor.physicalDimension();
      config.innerLines = dim.height / nfontHeight;
      config.innerChars = dim.width / nfontWidth;
      
      if (config.scrollbarVertical)
      {
         // set the horizontal scroll and location
         editorScroll.vbar().setMin(0);
         editorScroll.vbar().setMax(fontHeight * parsedText.size());
         
         int used = config.scrollY + config.cursorY;
         
         editorScroll.vbar().setPosition(used * fontHeight);
      }
      
      if (config.scrollbarHorizontal)
      {
         // set the horizontal scroll and location
         editorScroll.hbar().setMin(0);
         editorScroll.hbar().setMax(maxLineWidth);
         
         String txt = getLine().getText();
         txt = txt.substring(0, config.scrollX + config.cursorX);
         double used = gd.getTextWidth(txt, gf.font().font);
         used = screen().coordinates().widthFromNative((int) used);

         editorScroll.hbar().setPosition(used);
      }
   }

   /**
    * Compute the maximum line width in this editor, in character units.
    * 
    * @return   See above.
    */
   private double maxLineWidth()
   {
      CoordinatesConversion cc = screen().coordinates();

      double maxLineWidth = 0;
      for (Line line : parsedText)
      {
         int nlineWidth = gd.getTextWidth(line.getText(), gf.font().font);
         double lineWidth = cc.widthFromNative(nlineWidth);
         
         maxLineWidth = Math.max(lineWidth, maxLineWidth);
      }

      return maxLineWidth;
   }
   
   /**
    * Worker class which is responsible of drawing the editor's text.
    */
   private class EditorContent 
   extends AbstractContainer<GuiOutputManager>
   implements ScrollListener
   {

      @Override
      public Dimension dimension()
      {
         return minimumSize();
      }

      @Override
      public void draw()
      {
         final NativePoint p = physicalLocation();
         final NativeDimension d = physicalDimension();

         final NativeRectangle clip = clipRectangle(p, d);

         resizeScrollPane();
         gd.setGuiFont(gf.font().font);
         
         Line line = getLine();
         String txt = (line == null ? "" : line.getText().substring(0, config.scrollX));
         final int deltaScrollX = gd.getTextWidth(txt, gf.font().font);

         gd.draw(p, clip, new Runnable()
         {
            @Override
            public void run()
            {
               gd.setColor(gc.fgColor);
               
               int y = 0;
               int fontHeight = gf.font().fontHeight;
               
               int last = Math.min(config.scrollY + config.innerLines, parsedText.size());
               for (int line = config.scrollY; line < last; line++)
               {
                  Line l = parsedText.get(line);
                  // draw the texts, one by one
                  y += fontHeight;
                  String txt = l.getText();

                  // adjust for scrollX
                  gd.drawString(txt, -deltaScrollX, y);
               }
            }
         });
      }

      @Override
      public double width()
      {
         Point baseUnits = screen().coordinates().baseUnits();
         return parent().width() - baseUnits.x * 4;
      }

      @Override
      public double height()
      {
         Point baseUnits = screen().coordinates().baseUnits();
         return parent().height() - baseUnits.y * 4;
      }
      
      @Override
      public Dimension minimumSize()
      {
         return new Dimension(width(), height());
      }

      @Override
      public void scroll(ScrollEvent event)
      {
         /* switch (event.direction())
         {
            case LEFT:
               config.scrollX = Math.max(0, config.scrollX - 1);
               break;
            case RIGHT:
               break;
            case UP:
               config.scrollY = Math.max(0, config.scrollY - 1);
               break;
            case DOWN:
               int lastFirstRow = parsedText.size() - config.innerLines;
               config.scrollY = Math.min(lastFirstRow, config.scrollY + 1);
               break;
         }*/
      }
   }
}
