/* 
** Module   : LabelGuiImpl.java
** Abstract : Implementation for LABEL in GUI clients.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141220 Created initial version.
** 002 CA  20150323 Second pass at LABEL in GUI mode (side-labels currently).  The label's width 
**                  is not yet computed properly.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20150406 Fixed label drawing and size in GUI.
** 005 CA  20150507 Fixed label drawing - centered vertically.
** 006 CA  20150510 Set the label's font, before drawing the text.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.Font;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Implementation for a widget's LABEL, in a GUI client.
 */
public class LabelGuiImpl
extends Label<GuiOutputManager>
{
   /** Font resolver */
   private GuiFontResolver gf;

   /** Screen driver */
   private GuiDriver gd;

   /**
    * Create a copy of the label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   labelInstance
    *          Label to copy parameters from.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    */
   public LabelGuiImpl(WidgetId id, Label<GuiOutputManager> labelInstance, int frameId)
   {
      super(id, labelInstance, frameId);

      initialize(id, labelInstance.config());
   }

   /**
    * Constructs label with given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Initial text.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    */
   public LabelGuiImpl(WidgetId id, String text, int frameId)
   {
      super(id, text, frameId);
      
      initialize(id, config());
   }
   
   /**
    * Retrieve the widget's width in native units.
    *
    * TODO: this is duplicated in {@link FrameGuiImpl#getLabelWidth}, fix this.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeWidth()
   {
      if (gf == null)
      {
         return 0;
      }
      
      int font = gf.resolveFontNum();
      Window<?> window = Window.resolveWindow(this);
      
      int nativeWidth = FontManager.getTextWidth(name(), window, font, gd);
      
      return nativeWidth;
   }
   
   /**
    * Retrieve the widget's height in native units.
    * 
    * TODO: this is duplicated in {@link FrameGuiImpl#getLabelHeight}, fix this.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeHeight()
   {
      Window<?> window = Window.resolveWindow(this);
      int legacyHeight = 0;
      boolean text = true;
      // top-label case: the height is always the default-fixed-font's height!
      int font = FontManager.DEFAULT_FIXED_FONT;
      if (sideWidget != null)
      {
         // side-label case: use the same height as the widget's

         font = gf.resolveFontNum();
         text = (sideWidget instanceof Text);
      }
      
      legacyHeight = FontManager.getFontHeight(window, font, screen().getInstanceDriver());
      if (!text)
      {
         boolean threeD = FontManager.isThreeD();
         legacyHeight = legacyHeight + (int) (legacyHeight / 2) + (threeD ? 2 : 0);
      }

      return legacyHeight;
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, LabelConfig cfg)
   {
      super.initialize(id, cfg);

      gd = (GuiDriver) OutputManager.getDriver();
      
      // the label's window is resolved from the frame... the label is not necessarily attached
      // to the frame at the time of this call, just the frame ID is known

      Widget<?> parent = ThinClient.getInstance().getWidget(cfg.frameId);
      Window<?> window = parent.window();
      
      gf = new GuiFontResolver(window, gd, config(), false);

      updateSize();
   }

   /**
    * Compute the label's width.
    * 
    * @return   See above.
    */
   @Override
   public double width()
   {
      if (config().fixedWidth)
      {
         return config().widthChars;
      }
      
      return super.width();
   }
   
   /**
    * Compute the label's height.
    * 
    * @return   See above.
    */
   @Override
   public double height()
   {
      if (config().fixedHeight)
      {
         return config().heightChars;
      }
      
      return super.height();
   }

   /**
    * Draws text.
    */
   @Override
   public void draw()
   {
      if (parent() == null)
         return;

      FontCache fc = gf.font();
      final Font font = fc.font;

      final GuiColorResolver gce = new GuiColorResolver(gd, config(), 
                                                        SysColor.COLOR_FILLIN_BG, 
                                                        SysColor.COLOR_FILLIN_FG);
      gce.colors();
      
      NativePoint p = physicalLocation();
      final NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            gd.setColor(gce.bgColor);
            gd.fillRect(0, 0, dim.width, dim.height);

            if (!emptyMode)
            {
               String txt =  name();

               // needs to be centered vertically
               int txtHeight = FontManager.getTextHeight(txt, window(), gf.resolveFontNum(), screen().getInstanceDriver());
               // top labels are centered differently
               boolean centered = (sideWidget != null);
               int ytxt = (centered ? dim.height : dim.height * 2 - txtHeight);
   
               gd.setGuiFont(font);
               
               int xtxt = 1;
               if (rightAligned)
               {
                  xtxt = dim.width - gd.getTextWidth(txt, font);
               }
               
               gd.setColor(gce.fgColor);
               if (centered)
               {
                  gd.drawStringCentered(txt, xtxt, ytxt);
               }
               else
               {
                  gd.drawString(txt, xtxt, ytxt);
               }
            }
         }
      });
      
      if (isUnderlined())
      {
         // if underlined, draw a horizontal bar.  The width of this horizontal bar is
         // the label's width
         final int ppr = ThinClient.getInstance().getPixelsPerRow();
         
         final NativeDimension sepDim = new NativeDimension(dim.width, ppr);
         NativePoint sepLoc = new NativePoint(p.x, (isDelimiter() ? p.y : dim.height));
         NativeRectangle sepClip = clipRectangle(sepLoc, sepDim);

         gd.draw(sepLoc, sepClip, new Runnable()
         {
            @Override
            public void run()
            {
               int ysep = ppr / 2;
               gd.setColor(gce.fgColor);
               gd.drawLine(0, ysep, sepDim.width, ysep);
            }
         });
      }

      if (disableEmptyMode)
         emptyMode = false;
      
      disableEmptyMode = false;
   }
}
